from aioclustermanager.job_list import JobList
from aioclustermanager.nomad.job import NomadJob


class NomadJobList(JobList):

    def __init__(self, data=None, namespace='', **kw):
        self._raw = data
        self._jobs = []
        for job in self._raw:
            job_obj = NomadJob(data=job)
            job_obj._namespace = namespace
            self._jobs.append(job_obj)

    @property
    def total(self):
        return len(self._jobs)

    def values(self):
        return self._jobs



from copy import deepcopy

K8S_DELETE = {
    'kind': 'DeleteOptions',
    'propagationPolicy': 'Background'
}


class K8SDelete(object):

    def __init__(self):
        self._raw = K8S_DELETE

    def payload(self):
        return deepcopy(self._raw)



from aioclustermanager.k8s.caller import K8SCaller
from base64 import b64decode

import aiohttp
import logging
import os
import ssl
import tempfile

logger = logging.getLogger('aioclustermanager')


class Configuration:
    file = None

    def __init__(self, environment):
        self.environment = environment
        if environment['certificate'] is not None:
            # Certificate management
            ssl_client_context = ssl.create_default_context(
                purpose=ssl.Purpose.CLIENT_AUTH)
            self.cert_file = tempfile.NamedTemporaryFile(delete=False)
            self.cert_file.write(b64decode(environment['certificate']))
            self.cert_file.close()
            self.client_key = tempfile.NamedTemporaryFile(delete=False)
            self.client_key.write(b64decode(environment['key']))
            self.client_key.close()

            ssl_client_context.load_cert_chain(
                certfile=self.cert_file.name, keyfile=self.client_key.name)
            conn = aiohttp.TCPConnector(ssl_context=ssl_client_context)
            self.session = aiohttp.ClientSession(connector=conn)
        elif environment['certificate_file'] is not None:
            logger.debug('Loading cert files')
            ssl_client_context = ssl.create_default_context(
                purpose=ssl.Purpose.CLIENT_AUTH)
            ssl_client_context.load_cert_chain(
                certfile=environment['certificate_file'],
                keyfile=environment['key_file'])
            conn = aiohttp.TCPConnector(ssl_context=ssl_client_context)
            self.session = aiohttp.ClientSession(connector=conn)
        else:
            basic_auth = aiohttp.BasicAuth(
                environment['user'], environment['credentials'])
            self.session = aiohttp.ClientSession(auth=basic_auth)

        ssl_context = None
        if environment['ca'] is not None:
            self.file = tempfile.NamedTemporaryFile(delete=False)
            self.file.write(bytes(environment['ca'], encoding='utf-8'))
            self.file.close()
            self.ssl_context = ssl.SSLContext()
            ssl_context.load_verify_locations(self.file.name)
        elif environment['ca_file'] is not None:
            self.ssl_context = ssl.SSLContext()
            self.ssl_context.load_verify_locations(environment['ca_file'])

        if environment['skip_ssl']:
            self.verify = False
        else:
            self.verify = True


class K8SContextManager:
    def __init__(self, environment):
        self.environment = environment

    async def __aenter__(self):
        return await self.open()

    async def open(self):
        self.config = Configuration(self.environment)
        return K8SCaller(
            self.config.ssl_context,
            self.environment['endpoint'],
            self.config.session,
            verify=self.config.verify)

    async def __aexit__(self, exc_type, exc, tb):
        if self.config.file is not None:
            os.unlink(self.config.file.name)
        await self.config.session.close()


async def create_k8s_caller(environment):
    config = Configuration(environment)
    return K8SCaller(
        config.ssl_context,
        environment['endpoint'],
        config.session,
        verify=config.verify)



from aioclustermanager.nomad import const
from aioclustermanager.nomad.executions_list import NomadExecutionList
from aioclustermanager.nomad.job import NomadJob
from aioclustermanager.nomad.job_list import NomadJobList
from aioclustermanager.nomad.namespace import NomadNamespace

import asyncio
import logging

# from aioclustermanager.utils import convert

logger = logging.getLogger(__name__)

WATCH_OPS = {
    'job': 'http://{endpoint}/v1/job/{namespace}-{name}/allocations',
    'namespace': None
}

GET_OPS = {
    'list_jobs': 'http://{endpoint}/v1/jobs',
    'job': 'http://{endpoint}/v1/job/{namespace}-{name}',
    'executions': 'http://{endpoint}/v1/job/{namespace}-{name}/allocations'
}

POST_OPS = {
    'job': 'http://{endpoint}/v1/jobs'
}

DELETE_OPS = {
    'job': 'http://{endpoint}/v1/job/{namespace}-{name}?purge=true'
}


class NomadCaller:
    _type = 'nomad'
    constants = const

    def __init__(self, endpoint, datacenters, session):
        self.endpoint = endpoint
        self.session = session
        self.datacenters = datacenters

    @property
    def type(self):
        return self._type

    # Namespaces, implementation using prefix jobs

    async def get_namespace(self, namespace):
        return NomadNamespace(name=namespace)

    async def create_namespace(self, namespace):
        return True

    async def delete_namespace(self, namespace):
        jobs = await self.list_jobs(namespace)
        for job in jobs:
            await self.delete_job(namespace, job.id)
        return True

    async def wait_added(self, kind, namespace, name=None):
        url = WATCH_OPS[kind]
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        return await self.watch(url, value='Started')  # correct status?

    async def wait_deleted(self, kind, namespace, name=None):
        url = WATCH_OPS[kind]
        if url is None:
            return False
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        return await self.watch(url, value='Deleted')  # correct status?

    async def define_quota(self, namespace, cpu_limit=None, mem_limit=None):
        # Only supported on nomad enterprise
        pass

    async def get_job(self, namespace, name):
        url = GET_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        params = {}
        result = await self.get(url, params)
        if result is None:
            return None
        else:
            job_obj = NomadJob(data=result)
            job_obj._namespace = namespace
            return job_obj

    async def get_tfjob(self, namespace, name):
        # TODO
        return None

    async def get_job_executions(self, namespace, name):
        url = GET_OPS['executions']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        params = {}
        result = await self.get(url, params)
        if result is None:
            return None
        else:
            return NomadExecutionList(data=result)

    async def delete_job(self, namespace, name, wait=False):
        url = DELETE_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        await self.delete(url, None, None)
        if wait:
            return await self.wait_deleted('job', namespace, name)
        return True

    async def delete_tfjob(self, namespace, name, wait=False):
        # TODO
        return None

    async def list_jobs(self, namespace):
        url = GET_OPS['list_jobs']
        params = {
            'prefix': namespace + '-'
        }
        url = url.format(
            namespace=namespace,
            endpoint=self.endpoint)
        result = await self.get(url, params)
        if result is None:
            return None
        else:
            return NomadJobList(data=result, namespace=namespace)

    async def list_tfjobs(self, namespace):
        # TODO
        return []

    async def create_job(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}):
        url = POST_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        obj = NomadJob(
            namespace=namespace,
            name=name,
            image=image,
            command=command, args=args,
            cpu_limit=cpu_limit, mem_limit=mem_limit,
            envvars=envvars)
        obj._namespace = namespace
        obj.set_datacenters([self.datacenters])
        return await self.post(url, None, obj.payload())

    async def create_tfjob(
            self, namespace, name, image,
            command=None, args=None,
            cpu_limit=None, mem_limit=None,
            envvars={}, workers=1, ps=1,
            masters=1, tb_gs=None):
        # TODO
        return None

    async def wait_for_job(self, namespace, name):
        url = WATCH_OPS['job']
        url = url.format(
            namespace=namespace,
            name=name,
            endpoint=self.endpoint)
        await self.watch(url, 'Terminated')
        job = await self.get_job(namespace, name)
        if not job.finished:
            raise Exception('Not finished')
        executions = await self.get_job_executions(namespace, name)
        return 1 if executions.is_done() else 0

    # BASIC OPS

    async def watch(self, url, value=None, timeout=20):
        not_found = True
        # TODO add timeout
        while not_found:
            allocations = await self.get(url, {})
            for alloc in allocations:
                if alloc['TaskStates'] is not None:
                    for key, values in alloc['TaskStates'].items():
                        for event in values['Events']:
                            logger.debug("Status : " + event['Type'])
                            if event['Type'] in [value]:
                                not_found = False
                                break
                if not_found is False:
                    break
            if not_found is True:
                await asyncio.sleep(5)

    async def get(self, url, params):
        async with self.session.get(
                url,
                headers={
                    'Accept': 'application/json'
                },
                params=params) as resp:
            if resp.status == 404:
                return None
            assert resp.status == 200
            result = await resp.json()
            return result

    async def post(self, url, version, payload):
        async with self.session.post(
                url,
                json=payload,
                headers={
                    'Content-Type': 'application/json',
                    'Accept': 'application/json'
                }) as resp:
            if resp.status in [201, 200]:
                return await resp.json()
            else:
                text = await resp.text()
                logger.error('Error call: %d - %s' % (resp.status, text))
                raise Exception('Error calling nomad')

    async def delete(self, url, version, payload):
        async with self.session.delete(
                url,
                headers={
                    'Content-Type': 'application/json',
                    'Accept': 'application/json'
                }) as resp:
            if resp.status == 200:
                return await resp.json()
            elif resp.status == 404:
                return None
            else:
                text = await resp.text()
                logger.error('Error call: %d - %s' % (resp.status, text))
                raise Exception('Error calling nomad')



from copy import deepcopy

NOMAD_NAMESPACE = {
    "name": "",
}


class NomadNamespace:

    def __init__(self, name=None, data=None):
        if data is not None:
            self._raw = data
        else:
            self._raw = self.create(
                name)

    def create(self, name):
        namespace_info = deepcopy(NOMAD_NAMESPACE)
        namespace_info['name'] = name
        return namespace_info

    def payload(self):
        return deepcopy(self._raw)



from dateutil.parser import parse


class Execution:
    """Generic pod class."""

    def __init__(self, data=None):
        self._raw = data

    @property
    def id(self):
        raise NotImplementedError()

    @property
    def pending(self):
        raise NotImplementedError()

    @property
    def running(self):
        raise NotImplementedError()

    @property
    def status(self):
        raise NotImplementedError()

    @property
    def events(self):
        raise NotImplementedError()

    def get_payload(self):
        raise NotImplementedError()

    def payload(self):
        return self._raw

    def sort_status_executions(self):
        result = sorted(self.events, key=lambda x: parse(x[1]))
        return result



from aioclustermanager.job import Job
from copy import deepcopy

K8S_JOB = {
    "kind": "TFJob",
    "metadata": {
        "name": "",
        "namespace": ""
    },
    "spec": {
        "replicaSpecs": [{
            "replicas": 1,
            "tfReplicaType": "WORKER",
            "template": {
                "spec": {
                    "containers": [{
                        "image": "",
                        "name": "",
                        "resources": {
                            "limits": {
                            }
                        }
                    }],
                    # XXX are we sure we want to restart these jobs?
                    # what if they restart continuously on bad code?
                    "restartPolicy": "OnFailure"
                }
            }
        }, {
            "replicas": 1,
            "tfReplicaType": "MASTER",
            "template": {
                "spec": {
                    "containers": [{
                        "image": "",
                        "name": "",
                        "resources": {
                            "limits": {
                            }
                        }
                    }],
                    "restartPolicy": "OnFailure"
                }
            }
        }, {
            "replicas": 1,
            "tfReplicaType": "PS"
        }]
    }
}


class K8STFJob(Job):
    @property
    def active(self):
        return self._raw['status']['active']

    @property
    def finished(self):
        status = self._raw['status']
        return 'failed' in status or 'succeeded' in status

    @property
    def id(self):
        return self._raw['metadata']['name']

    def create(self, namespace, name, image, **kw):
        job_info = deepcopy(K8S_JOB)
        job_info['metadata']['name'] = name
        job_info['metadata']['namespace'] = namespace
        # have to love the nesting here...
        job_info['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]['name'] = name  # noqa
        job_info['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]['image'] = image  # noqa
        job_info['spec']['replicaSpecs'][1]['template']['spec']['containers'][0]['name'] = name  # noqa
        job_info['spec']['replicaSpecs'][1]['template']['spec']['containers'][0]['image'] = image  # noqa

        if 'command' in kw:
            job_info['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]['command'] = kw['command']  # noqa
            job_info['spec']['replicaSpecs'][1]['template']['spec']['containers'][0]['command'] = kw['command']  # noqa

        if 'args' in kw:
            job_info['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]['args'] = kw['args']  # noqa
            job_info['spec']['replicaSpecs'][1]['template']['spec']['containers'][0]['args'] = kw['args']  # noqa

        if 'mem_limit' in kw:
            job_info['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]['resources']['limits']['memory'] = kw['mem_limit']  # noqa
            job_info['spec']['replicaSpecs'][1]['template']['spec']['containers'][0]['resources']['limits']['memory'] = kw['mem_limit']  # noqa

        if 'cpu_limit' in kw:
            job_info['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]['resources']['limits']['cpu'] = kw['cpu_limit']  # noqa
            job_info['spec']['replicaSpecs'][1]['template']['spec']['containers'][0]['resources']['limits']['cpu'] = kw['cpu_limit']  # noqa

        if 'envs' in kw:
            envlist = []
            for key, value in kw['envs']:
                envlist.append({
                    "name": key,
                    "value": value
                })
            job_info['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]['env'] = envlist  # noqa
            job_info['spec']['replicaSpecs'][1]['template']['spec']['containers'][0]['env'] = envlist  # noqa

        return job_info

    def get_payload(self):
        container = self._raw['spec']['replicaSpecs'][0]['template']['spec']['containers'][0]  # noqa
        for env in container['env']:
            if env['name'] == 'PAYLOAD':
                data = env['value']
                return data
        return None



from aioclustermanager.utils import generate_word
from copy import deepcopy

K8S_LIMITS = {
    "kind": "ResourceQuota",
    "metadata": {
        "name": "",
        "namespace": ""
    },
    "spec": {
        "hard": {
            "requests.cpu": "",
            "requests.memory": "",
            "limits.cpu": "",
            "limits.memory": ""
        }
    }
}


class K8SQuota(object):

    def __init__(
            self, namespace=None, name=None, max_cpu=None,
            max_memory=None, data=None):
        if data is not None:
            self._raw = data
        else:
            if name is None:
                name = generate_word(10)
            self._raw = self.create(
                namespace,
                name=name,
                max_memory=max_memory,
                max_cpu=max_cpu)

    @property
    def id(self):
        return self._raw['metadata']['name']

    def create(self, namespace, name, max_memory, max_cpu):
        limit_info = deepcopy(K8S_LIMITS)
        limit_info['metadata']['name'] = name
        limit_info['metadata']['namespace'] = namespace
        limit_info['spec']['hard']['requests.cpu'] = max_cpu
        limit_info['spec']['hard']['requests.memory'] = max_memory
        limit_info['spec']['hard']['limits.cpu'] = max_cpu
        limit_info['spec']['hard']['limits.memory'] = max_memory
        return limit_info

    def payload(self):
        return deepcopy(self._raw)



from aioclustermanager.job import Job
from aioclustermanager.nomad.const import DEAD
from copy import deepcopy

NOMAD_JOB = {
    "Job": {
        "Constraints": [],
        "ID": "",
        "Name": "",
        "TaskGroups": [{
            "Constraints": None,
            "Name": "aioclustermanager",
            "Tasks": [{
                "Config": {
                    "image": ""
                },
                "Constraints": [],
                "Driver": "docker",
                "Env": {},
                "LogConfig": {
                    "MaxFileSizeMB": 10,
                    "MaxFiles": 10
                },
                "Name": "",
                "Resources": {
                    "CPU": 200,
                    "MemoryMB": 300
                },
                "Services": [],
                "Templates": [],
                "User": "",
            }]
        }],
        "Type": "batch",
    }
}


class NomadJob(Job):

    @property
    def namespace(self):
        return getattr(self, '_namespace', '')

    @property
    def active(self):
        return self._raw['Status'] != DEAD

    @property
    def finished(self):
        return self._raw['Status'] == DEAD

    @property
    def status(self):
        return self._raw['Status']

    @property
    def id(self):
        if self.namespace == '':
            return self._raw['ID']
        elif self._raw['ID'].startswith(self._namespace):
            to_remove = len(self._namespace) + 1
            return self._raw['ID'][to_remove:]

    @property
    def command(self):
        return self._raw['Job']['TaskGroups'][0]['Tasks'][0]['Config']['command']  # noqa

    @property
    def image(self):
        return self._raw['Job']['TaskGroups'][0]['Tasks'][0]['Config']['image']  # noqa

    # def get_payload(self):
    #     if 'TaskGroups' not in self._raw:
    #         return
    #     try:
    #         return json.loads(
    #             self._raw['TaskGroups'][0]['Tasks'][0]['Env']['PAYLOAD'])
    #     except (TypeError, KeyError, json.decoder.JSONDecodeError):
    #         try:
    #             task = self._raw['TaskGroups'][0]['Tasks'][0]
    #             templates = task['Templates']
    #             template = templates[-1]['EmbeddedTmpl']
    #             payload = template.split(
    #                 'cat <<EOF')[-1].split('EOF')[0].strip()
    #             return json.loads(payload)
    #         except (TypeError, AttributeError, KeyError,
    #                 json.decoder.JSONDecodeError):
    #             pass

    def set_datacenters(self, datacenters):
        self._raw['Job']['Datacenters'] = datacenters

    def create(self, namespace, name, image, **kw):
        job_info = deepcopy(NOMAD_JOB)

        real_job_id = '{}-{}'.format(namespace, name)

        job_info['Job']['ID'] = real_job_id
        job_info['Job']['Name'] = real_job_id
        job_info['Job']['TaskGroups'][0]['Tasks'][0]['Name'] = real_job_id

        job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['image'] = image

        if 'nomad_constraints' in kw and kw['nomad_constraints'] is not None:
            job_info['Job']['Constraints'] = kw['nomad_constraints']

        if 'docker_network_mode' in kw and kw['docker_network_mode'] is not None:  # noqa
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['network_mode'] = kw['docker_network_mode']  # noqa

        if 'volumes' in kw and kw['volumes'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['volumes'] = kw['volumes']  # noqa

        extend_args = []
        if 'command' in kw and kw['command'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['command'] = kw['command'][0]  # noqa

            if len(kw['command']) > 1:
                extend_args = kw['command'][1:]

        if 'args' in kw and kw['args'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Config']['args'] = ' '.join(extend_args) + ' ' + kw['args']  # noqa

        if 'templates' in kw and kw['templates'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Templates'] = kw['templates']  # noqa

        if 'cpu_limit' in kw and kw['cpu_limit'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Resources']['CPU'] = kw['cpu_limit']  # noqa

        if 'mem_limit' in kw and kw['mem_limit'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Resources']['MemoryMB'] = kw['mem_limit']  # noqa

        if 'envs' in kw and kw['envs'] is not None:
            job_info['Job']['TaskGroups'][0]['Tasks'][0]['Env'] = kw['envs']  # noqa

        if 'datacenters' in kw and kw['datacenters'] is not None:
            job_info['Job']['Datacenters'] = kw['datacenters']

        return job_info



from aioclustermanager.executions_list import ExecutionList
from aioclustermanager.k8s.const import FAILED, RUNNING, SUCCEEDED
from aioclustermanager.k8s.execution import K8SExecution


class K8SExecutionList(ExecutionList):

    def process(self):
        for pod in self._raw['items']:
            self._executions.append(K8SExecution(data=pod))

    def statuses(self):
        return [execution.status for execution in self]

    def has_failed_anytime(self):
        for execution in self:
            if execution.status == FAILED:
                return True
        return False

    def is_running(self):
        for execution in self:
            if execution.status == RUNNING:
                return True
        return False

    def is_done(self):
        for execution in self:
            if execution.status == SUCCEEDED:
                return True
        return False



import string
import random

VOWELS = "aeiou"
CONSONANTS = "".join(set(string.ascii_lowercase) - set(VOWELS))


def generate_word(length):
    word = ""
    for i in range(length):
        if i % 2 == 0:
            word += random.choice(CONSONANTS)
        else:
            word += random.choice(VOWELS)
    return word



from aioclustermanager.nomad.caller import NomadCaller

import aiohttp


class NomadContextManager:
    def __init__(self, environment):
        self.environment = environment
        self.session = None
        self._datacenter = None

    async def __aenter__(self):
        self.session = aiohttp.ClientSession()

        url = 'http://{}/v1/agent/self'.format(self.environment['endpoint'])
        async with self.session.get(url) as resp:
            data = await resp.json()
            self._datacenter = data['config']['Datacenter']

        return NomadCaller(
            self.environment['endpoint'],
            self._datacenter,
            self.session)

    async def __aexit__(self, exc_type, exc, tb):
        await self.session.close()



import pytest
pytestmark = pytest.mark.asyncio


async def test_get_jobs_nomad(nomad):
    # We clean up all the jobs on the namespace
    result = await nomad.cleanup_jobs('aiocluster-test')
    assert result == 0

    jobs_info = await nomad.list_jobs('aiocluster-test')
    assert jobs_info.total == 0

    await nomad.create_job(
        'aiocluster-test',
        'test-job',
        'perl',
        command=["perl", "-Mbignum=bpi", "-wle", "print bpi(2000)"])

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info.id == 'test-job'

    jobs_info = await nomad.list_jobs('aiocluster-test')
    assert jobs_info.total == 1

    result = await nomad.wait_for_job('aiocluster-test', 'test-job')
    assert result == 1

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info.finished

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info.id == 'test-job'

    result = await nomad.delete_job('aiocluster-test', 'test-job')
    assert result is True

    job_info = await nomad.get_job('aiocluster-test', 'test-job')
    assert job_info is None





class ExecutionList:
    """Generic job class."""

    def __init__(self, data=None, **kw):
        self._raw = data
        self._executions = []
        self.process()

    def process(self):
        raise NotImplementedError()

    def values(self):
        return self._executions

    def __len__(self):
        return len(self._executions)

    def __iter__(self):
        for execution in self._executions:
            yield execution

    def __setitem__(self, key, value):
        self._executions[key] = value

    def __getitem__(self, key):
        return self._executions[key]
